from langfuse.api.resources.commons.types.create_generation_request import CreateGenerationRequest
from langfuse.api.resources.commons.types.create_event_request import CreateEventRequest
from langfuse.api.resources.commons.types.create_span_request import CreateSpanRequest
from langfuse.api.resources.commons.types.llm_usage import LlmUsage
from langfuse.api.resources.generations.types.update_generation_request import UpdateGenerationRequest
from langfuse.api.resources.score.types.create_score_request import CreateScoreRequest

from langfuse.api.resources.span.types.update_span_request import UpdateSpanRequest
from langfuse.api.resources.trace.types.create_trace_request import CreateTraceRequest

# these imports need to stay here, otherwise imports from our clients wont work
from langfuse.api.resources.commons.types.trace_id_type_enum import TraceIdTypeEnum
from langfuse.api.resources.commons.types.observation_level import ObservationLevel


class InitialGeneration(CreateGenerationRequest):
    __fields__ = {name: field for name, field in CreateGenerationRequest.__fields__.items() if name not in ["traceIdType"]}


class InitialScore(CreateScoreRequest):
    __fields__ = {name: field for name, field in CreateScoreRequest.__fields__.items() if name not in ["traceIdType"]}


class InitialSpan(CreateSpanRequest):
    __fields__ = {name: field for name, field in CreateSpanRequest.__fields__.items() if name not in ["traceIdType"]}


class CreateScore(CreateScoreRequest):
    __fields__ = {name: field for name, field in CreateScoreRequest.__fields__.items() if name not in ["trace_id", "observation_id", "traceIdType"]}


class CreateTrace(CreateTraceRequest):
    __fields__ = {name: field for name, field in CreateTraceRequest.__fields__.items() if name not in ["release", "external_id"]}


class CreateGeneration(CreateGenerationRequest):
    __fields__ = {name: field for name, field in CreateGenerationRequest.__fields__.items() if name not in ["parent_observation_id", "traceIdType"]}


class CreateSpan(CreateSpanRequest):
    __fields__ = {name: field for name, field in CreateSpanRequest.__fields__.items() if name not in ["parent_observation_id", "traceIdType"]}


class CreateEvent(CreateEventRequest):
    __fields__ = {name: field for name, field in CreateEventRequest.__fields__.items() if name not in ["parent_observation_id", "traceIdType"]}


class UpdateGeneration(UpdateGenerationRequest):
    __fields__ = {name: field for name, field in UpdateGenerationRequest.__fields__.items() if name not in ["generation_id"]}


class UpdateSpan(UpdateSpanRequest):
    __fields__ = {name: field for name, field in UpdateSpanRequest.__fields__.items() if name not in ["spanId", "span_id"]}


class Usage(LlmUsage):
    __fields__ = {name: field for name, field in LlmUsage.__fields__.items()}



# This file was auto-generated by Fern from our API Definition.

import typing

from .resources.event.client import AsyncEventClient, EventClient
from .resources.generations.client import AsyncGenerationsClient, GenerationsClient
from .resources.observations.client import AsyncObservationsClient, ObservationsClient
from .resources.score.client import AsyncScoreClient, ScoreClient
from .resources.span.client import AsyncSpanClient, SpanClient
from .resources.trace.client import AsyncTraceClient, TraceClient


class FintoLangfuse:
    def __init__(
        self,
        *,
        environment: str,
        x_langfuse_sdk_name: typing.Optional[str] = None,
        x_langfuse_sdk_version: typing.Optional[str] = None,
        username: str,
        password: str
    ):
        self._environment = environment
        self.x_langfuse_sdk_name = x_langfuse_sdk_name
        self.x_langfuse_sdk_version = x_langfuse_sdk_version
        self._username = username
        self._password = password
        self.event = EventClient(
            environment=self._environment,
            x_langfuse_sdk_name=self.x_langfuse_sdk_name,
            x_langfuse_sdk_version=self.x_langfuse_sdk_version,
            username=self._username,
            password=self._password,
        )
        self.generations = GenerationsClient(
            environment=self._environment,
            x_langfuse_sdk_name=self.x_langfuse_sdk_name,
            x_langfuse_sdk_version=self.x_langfuse_sdk_version,
            username=self._username,
            password=self._password,
        )
        self.observations = ObservationsClient(
            environment=self._environment,
            x_langfuse_sdk_name=self.x_langfuse_sdk_name,
            x_langfuse_sdk_version=self.x_langfuse_sdk_version,
            username=self._username,
            password=self._password,
        )
        self.score = ScoreClient(
            environment=self._environment,
            x_langfuse_sdk_name=self.x_langfuse_sdk_name,
            x_langfuse_sdk_version=self.x_langfuse_sdk_version,
            username=self._username,
            password=self._password,
        )
        self.span = SpanClient(
            environment=self._environment,
            x_langfuse_sdk_name=self.x_langfuse_sdk_name,
            x_langfuse_sdk_version=self.x_langfuse_sdk_version,
            username=self._username,
            password=self._password,
        )
        self.trace = TraceClient(
            environment=self._environment,
            x_langfuse_sdk_name=self.x_langfuse_sdk_name,
            x_langfuse_sdk_version=self.x_langfuse_sdk_version,
            username=self._username,
            password=self._password,
        )


class AsyncFintoLangfuse:
    def __init__(
        self,
        *,
        environment: str,
        x_langfuse_sdk_name: typing.Optional[str] = None,
        x_langfuse_sdk_version: typing.Optional[str] = None,
        username: str,
        password: str
    ):
        self._environment = environment
        self.x_langfuse_sdk_name = x_langfuse_sdk_name
        self.x_langfuse_sdk_version = x_langfuse_sdk_version
        self._username = username
        self._password = password
        self.event = AsyncEventClient(
            environment=self._environment,
            x_langfuse_sdk_name=self.x_langfuse_sdk_name,
            x_langfuse_sdk_version=self.x_langfuse_sdk_version,
            username=self._username,
            password=self._password,
        )
        self.generations = AsyncGenerationsClient(
            environment=self._environment,
            x_langfuse_sdk_name=self.x_langfuse_sdk_name,
            x_langfuse_sdk_version=self.x_langfuse_sdk_version,
            username=self._username,
            password=self._password,
        )
        self.observations = AsyncObservationsClient(
            environment=self._environment,
            x_langfuse_sdk_name=self.x_langfuse_sdk_name,
            x_langfuse_sdk_version=self.x_langfuse_sdk_version,
            username=self._username,
            password=self._password,
        )
        self.score = AsyncScoreClient(
            environment=self._environment,
            x_langfuse_sdk_name=self.x_langfuse_sdk_name,
            x_langfuse_sdk_version=self.x_langfuse_sdk_version,
            username=self._username,
            password=self._password,
        )
        self.span = AsyncSpanClient(
            environment=self._environment,
            x_langfuse_sdk_name=self.x_langfuse_sdk_name,
            x_langfuse_sdk_version=self.x_langfuse_sdk_version,
            username=self._username,
            password=self._password,
        )
        self.trace = AsyncTraceClient(
            environment=self._environment,
            x_langfuse_sdk_name=self.x_langfuse_sdk_name,
            x_langfuse_sdk_version=self.x_langfuse_sdk_version,
            username=self._username,
            password=self._password,
        )



# This file was auto-generated by Fern from our API Definition.

from .create_event_request import CreateEventRequest
from .create_generation_request import CreateGenerationRequest
from .create_span_request import CreateSpanRequest
from .llm_usage import LlmUsage
from .map_value import MapValue
from .observation import Observation
from .observation_level import ObservationLevel
from .score import Score
from .trace import Trace
from .trace_id_type_enum import TraceIdTypeEnum
from .trace_with_details import TraceWithDetails
from .trace_with_full_details import TraceWithFullDetails

__all__ = [
    "CreateEventRequest",
    "CreateGenerationRequest",
    "CreateSpanRequest",
    "LlmUsage",
    "MapValue",
    "Observation",
    "ObservationLevel",
    "Score",
    "Trace",
    "TraceIdTypeEnum",
    "TraceWithDetails",
    "TraceWithFullDetails",
]



# This file was auto-generated by Fern from our API Definition.

from .access_denied_error import AccessDeniedError
from .error import Error
from .method_not_allowed_error import MethodNotAllowedError
from .not_found_error import NotFoundError
from .unauthorized_error import UnauthorizedError

__all__ = ["AccessDeniedError", "Error", "MethodNotAllowedError", "NotFoundError", "UnauthorizedError"]



# This file was auto-generated by Fern from our API Definition.

import enum
import typing

T_Result = typing.TypeVar("T_Result")


class ObservationLevel(str, enum.Enum):
    DEBUG = "DEBUG"
    DEFAULT = "DEFAULT"
    WARNING = "WARNING"
    ERROR = "ERROR"

    def visit(
        self,
        debug: typing.Callable[[], T_Result],
        default: typing.Callable[[], T_Result],
        warning: typing.Callable[[], T_Result],
        error: typing.Callable[[], T_Result],
    ) -> T_Result:
        if self is ObservationLevel.DEBUG:
            return debug()
        if self is ObservationLevel.DEFAULT:
            return default()
        if self is ObservationLevel.WARNING:
            return warning()
        if self is ObservationLevel.ERROR:
            return error()



# This file was auto-generated by Fern from our API Definition.

from . import commons, event, generations, observations, score, span, trace, utils
from .commons import (
    AccessDeniedError,
    CreateEventRequest,
    CreateGenerationRequest,
    CreateSpanRequest,
    Error,
    LlmUsage,
    MapValue,
    MethodNotAllowedError,
    NotFoundError,
    Observation,
    ObservationLevel,
    Score,
    Trace,
    TraceIdTypeEnum,
    TraceWithDetails,
    TraceWithFullDetails,
    UnauthorizedError,
)
from .generations import UpdateGenerationRequest
from .score import CreateScoreRequest, Scores
from .span import UpdateSpanRequest
from .trace import CreateTraceRequest, Traces

__all__ = [
    "AccessDeniedError",
    "CreateEventRequest",
    "CreateGenerationRequest",
    "CreateScoreRequest",
    "CreateSpanRequest",
    "CreateTraceRequest",
    "Error",
    "LlmUsage",
    "MapValue",
    "MethodNotAllowedError",
    "NotFoundError",
    "Observation",
    "ObservationLevel",
    "Score",
    "Scores",
    "Trace",
    "TraceIdTypeEnum",
    "TraceWithDetails",
    "TraceWithFullDetails",
    "Traces",
    "UnauthorizedError",
    "UpdateGenerationRequest",
    "UpdateSpanRequest",
    "commons",
    "event",
    "generations",
    "observations",
    "score",
    "span",
    "trace",
    "utils",
]



# This file was auto-generated by Fern from our API Definition.

from .errors import AccessDeniedError, Error, MethodNotAllowedError, NotFoundError, UnauthorizedError
from .types import (
    CreateEventRequest,
    CreateGenerationRequest,
    CreateSpanRequest,
    LlmUsage,
    MapValue,
    Observation,
    ObservationLevel,
    Score,
    Trace,
    TraceIdTypeEnum,
    TraceWithDetails,
    TraceWithFullDetails,
)

__all__ = [
    "AccessDeniedError",
    "CreateEventRequest",
    "CreateGenerationRequest",
    "CreateSpanRequest",
    "Error",
    "LlmUsage",
    "MapValue",
    "MethodNotAllowedError",
    "NotFoundError",
    "Observation",
    "ObservationLevel",
    "Score",
    "Trace",
    "TraceIdTypeEnum",
    "TraceWithDetails",
    "TraceWithFullDetails",
    "UnauthorizedError",
]



# This file was auto-generated by Fern from our API Definition.

import datetime as dt
import typing

import pydantic

from ....core.datetime_utils import serialize_datetime
from ...commons.types.llm_usage import LlmUsage
from ...commons.types.map_value import MapValue
from ...commons.types.observation_level import ObservationLevel


class UpdateGenerationRequest(pydantic.BaseModel):
    generation_id: str = pydantic.Field(alias="generationId")
    trace_id: typing.Optional[str] = pydantic.Field(alias="traceId")
    name: typing.Optional[str]
    end_time: typing.Optional[dt.datetime] = pydantic.Field(alias="endTime")
    completion_start_time: typing.Optional[dt.datetime] = pydantic.Field(alias="completionStartTime")
    model: typing.Optional[str]
    model_parameters: typing.Optional[typing.Dict[str, MapValue]] = pydantic.Field(alias="modelParameters")
    prompt: typing.Optional[typing.Any]
    version: typing.Optional[str]
    metadata: typing.Optional[typing.Any]
    completion: typing.Optional[str]
    usage: typing.Optional[LlmUsage]
    level: typing.Optional[ObservationLevel]
    status_message: typing.Optional[str] = pydantic.Field(alias="statusMessage")

    def json(self, **kwargs: typing.Any) -> str:
        kwargs_with_defaults: typing.Any = {"by_alias": True, "exclude_unset": True, **kwargs}
        return super().json(**kwargs_with_defaults)

    def dict(self, **kwargs: typing.Any) -> typing.Dict[str, typing.Any]:
        kwargs_with_defaults: typing.Any = {"by_alias": True, "exclude_unset": True, **kwargs}
        return super().dict(**kwargs_with_defaults)

    class Config:
        frozen = True
        allow_population_by_field_name = True
        json_encoders = {dt.datetime: serialize_datetime}



from datetime import datetime
import logging
from typing import Any, Dict, List, Optional, Sequence, Union
from uuid import UUID
from langchain.callbacks.base import BaseCallbackHandler
from langfuse.api.resources.commons.types.llm_usage import LlmUsage
from langfuse.api.resources.commons.types.observation_level import ObservationLevel
from langfuse.model import CreateGeneration, CreateSpan, CreateTrace, UpdateGeneration, UpdateSpan
from langchain.schema.output import LLMResult
from langchain.schema.messages import BaseMessage
from langchain.schema.document import Document
from langfuse.client import Langfuse, StatefulClient
from langchain.schema.agent import AgentAction, AgentFinish


class Run:
    def __init__(self, state: StatefulClient, parent_id: Optional[str]) -> None:
        self.state = state
        self.parent_id = parent_id


class CallbackHandler(BaseCallbackHandler):
    def __init__(self, public_key: str, secret_key: str, host: Optional[str]) -> None:
        self.langfuse = Langfuse(public_key, secret_key, host)
        self.trace = None
        self.runs = {}

    def on_chain_start(
        self,
        serialized: Dict[str, Any],
        inputs: Dict[str, Any],
        *,
        run_id: UUID,
        parent_run_id: UUID | None = None,
        tags: List[str] | None = None,
        metadata: Dict[str, Any] | None = None,
        **kwargs: Any,
    ):
        try:
            logging.debug(f"on chain start: {run_id}")
            self.__generate_trace_and_parent(serialized=serialized, inputs=inputs, run_id=run_id, parent_run_id=parent_run_id, tags=tags, metadata=metadata, kwargs=kwargs)
        except Exception as e:
            logging.error(e)

    def get_trace_id(self) -> str:
        return self.trace.state.id

    def __generate_trace_and_parent(
        self,
        serialized: Dict[str, Any],
        inputs: Dict[str, Any],
        *,
        run_id: UUID,
        parent_run_id: UUID | None = None,
        tags: List[str] | None = None,
        metadata: Dict[str, Any] | None = None,
        **kwargs: Any,
    ):
        try:
            class_name = serialized.get("name", serialized.get("id", ["<unknown>"])[-1])

            if self.trace is None:
                trace = Run(
                    self.langfuse.trace(
                        CreateTrace(
                            name=class_name,
                            metadata=self.__join_tags_and_metadata(tags, metadata),
                        )
                    ),
                    None,
                )
                self.trace = trace

            if parent_run_id is not None and parent_run_id in self.runs:
                self.runs[run_id] = Run(
                    self.runs[parent_run_id].state.span(
                        CreateSpan(
                            name=class_name,
                            metadata=self.__join_tags_and_metadata(tags, metadata),
                            input=inputs,
                            startTime=datetime.now(),
                        )
                    ),
                    parent_run_id,
                )
            else:
                self.runs[run_id] = Run(
                    self.trace.state.span(
                        CreateSpan(
                            name=class_name,
                            metadata=self.__join_tags_and_metadata(tags, metadata),
                            input=inputs,
                            startTime=datetime.now(),
                        )
                    ),
                    parent_run_id,
                )

        except Exception as e:
            logging.error(e)

    def on_agent_action(
        self,
        action: AgentAction,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        """Run on agent action."""
        try:
            logging.debug(f"on agent action: {run_id}")

            if run_id not in self.runs:
                raise Exception("run not found")

            self.runs[run_id].state = self.runs[run_id].state.update(UpdateSpan(endTime=datetime.now(), output=action))
        except Exception as e:
            logging.error(e)

    def on_agent_finish(
        self,
        finish: AgentFinish,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        try:
            logging.debug(f"on agent finish: {run_id}")
            if run_id not in self.runs:
                raise Exception("run not found")

            self.runs[run_id].state = self.runs[run_id].state.update(UpdateSpan(endTime=datetime.now(), output=finish))
        except Exception as e:
            logging.error(e)

    def on_chain_end(
        self,
        outputs: Dict[str, Any],
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        try:
            logging.debug(f"on chain end: {run_id}")

            if run_id not in self.runs:
                raise Exception("run not found")

            self.runs[run_id].state = self.runs[run_id].state.update(UpdateSpan(output=outputs, endTime=datetime.now()))
        except Exception as e:
            logging.error(e)

    def on_chain_error(
        self,
        error: Union[Exception, KeyboardInterrupt],
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> None:
        try:
            logging.debug(f"on chain error: {run_id}")
            self.runs[run_id].state = self.runs[run_id].state.update(UpdateSpan(level=ObservationLevel.ERROR, statusMessage=str(error), endTime=datetime.now()))
        except Exception as e:
            logging.error(e)

    def on_chat_model_start(
        self,
        serialized: Dict[str, Any],
        messages: List[List[BaseMessage]],
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[List[str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Any:
        try:
            logging.debug(f"on chat model start: {run_id}")
            self.__on_llm_action(serialized, run_id, messages, parent_run_id, tags=tags, metadata=metadata, **kwargs)
        except Exception as e:
            logging.error(e)

    def on_llm_start(
        self,
        serialized: Dict[str, Any],
        prompts: List[str],
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[List[str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Any:
        try:
            logging.debug(f"on llm start: {run_id}")
            self.__on_llm_action(serialized, run_id, prompts, parent_run_id, tags=tags, metadata=metadata, **kwargs)
        except Exception as e:
            logging.error(e)

    def on_tool_start(
        self,
        serialized: Dict[str, Any],
        input_str: str,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[List[str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Any:
        try:
            logging.debug(f"on tool start: {run_id}")

            if parent_run_id is None or parent_run_id not in self.runs:
                raise Exception("parent run not found")
            meta = self.__join_tags_and_metadata(tags, metadata)

            meta.update({key: value for key, value in kwargs.items() if value is not None})

            self.runs[run_id] = Run(
                self.runs[parent_run_id].state.span(
                    CreateSpan(
                        name=serialized.get("name", serialized.get("id", ["<unknown>"])[-1]),
                        input=input_str,
                        startTime=datetime.now(),
                        metadata=meta,
                    )
                ),
                parent_run_id,
            )
        except Exception as e:
            logging.error(e)

    def on_retriever_start(
        self,
        serialized: Dict[str, Any],
        query: str,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[List[str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Any:
        try:
            logging.debug(f"on retriever start: {run_id}")

            if parent_run_id is None or parent_run_id not in self.runs:
                raise Exception("parent run not found")

            self.runs[run_id] = Run(
                self.runs[parent_run_id].state.span(
                    CreateSpan(
                        name=serialized.get("name", serialized.get("id", ["<unknown>"])[-1]),
                        input=query,
                        startTime=datetime.now(),
                        metadata=self.__join_tags_and_metadata(tags, metadata),
                    )
                ),
                parent_run_id,
            )
        except Exception as e:
            logging.error(e)

    def on_retriever_end(
        self,
        documents: Sequence[Document],
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        try:
            logging.debug(f"on retriever end: {run_id}")

            if run_id is None or run_id not in self.runs:
                raise Exception("run not found")

            self.runs[run_id].state = self.runs[run_id].state.update(UpdateSpan(output=documents, endTime=datetime.now()))
        except Exception as e:
            logging.error(e)

    def on_tool_end(
        self,
        output: str,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        try:
            logging.debug(f"on tool end: {run_id}")
            if run_id is None or run_id not in self.runs:
                raise Exception("run not found")

            self.runs[run_id].state = self.runs[run_id].state.update(UpdateSpan(output=output, endTime=datetime.now()))
        except Exception as e:
            logging.error(e)

    def on_tool_error(
        self,
        error: Union[Exception, KeyboardInterrupt],
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        try:
            logging.debug(f"on tool error: {run_id}")
            if run_id is None or run_id not in self.runs:
                raise Exception("run not found")

            self.runs[run_id].state = self.runs[run_id].state.update(UpdateSpan(statusMessage=error, level=ObservationLevel.ERROR, endTime=datetime.now()))
        except Exception as e:
            logging.error(e)

    def __on_llm_action(
        self,
        serialized: Dict[str, Any],
        run_id: UUID,
        prompts: List[str],
        parent_run_id: Optional[UUID] = None,
        tags: Optional[List[str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ):
        try:
            if self.trace is None:
                # simple LLM call that has no trace and parent
                self.__generate_trace_and_parent(serialized, inputs=prompts, run_id=run_id, parent_run_id=parent_run_id, tags=tags, metadata=metadata, kwargs=kwargs)

            self.runs[run_id] = Run(
                self.runs[parent_run_id].state.generation(
                    CreateGeneration(
                        name=serialized.get("name", serialized.get("id", ["<unknown>"])[-1]),
                        prompt=prompts,
                        startTime=datetime.now(),
                        metadata=self.__join_tags_and_metadata(tags, metadata),
                        model=kwargs["invocation_params"]["model_name"],
                        modelParameters={
                            key: value
                            for key, value in {
                                "temperature": kwargs["invocation_params"].get("temperature"),
                                "max_tokens": kwargs["invocation_params"].get("max_tokens"),
                                "top_p": kwargs["invocation_params"].get("top_p"),
                                "frequency_penalty": kwargs["invocation_params"].get("frequency_penalty"),
                                "presence_penalty": kwargs["invocation_params"].get("presence_penalty"),
                                "request_timeout": kwargs["invocation_params"].get("request_timeout"),
                            }.items()
                            if value is not None
                        },
                    )
                )
                if parent_run_id in self.runs
                else self.trace.state.generation(
                    CreateGeneration(
                        name=serialized.get("name", serialized.get("id", ["<unknown>"])[-1]),
                        prompt=prompts,
                        startTime=datetime.now(),
                        metadata=self.__join_tags_and_metadata(tags, metadata),
                        model=kwargs["invocation_params"]["model_name"],
                        modelParameters={
                            key: value
                            for key, value in {
                                "temperature": kwargs["invocation_params"].get("temperature"),
                                "max_tokens": kwargs["invocation_params"].get("max_tokens"),
                                "top_p": kwargs["invocation_params"].get("top_p"),
                                "frequency_penalty": kwargs["invocation_params"].get("frequency_penalty"),
                                "presence_penalty": kwargs["invocation_params"].get("presence_penalty"),
                                "request_timeout": kwargs["invocation_params"].get("request_timeout"),
                            }.items()
                            if value is not None
                        },
                    )
                ),
                datetime.now(),
            )
        except Exception as e:
            logging.error(e)

    def on_llm_end(
        self,
        response: LLMResult,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        try:
            logging.debug(f"on llm end: {run_id}")
            if run_id not in self.runs:
                raise Exception("run not found")
            else:
                last_response = response.generations[-1][-1].text
                self.runs[run_id].state = self.runs[run_id].state.update(UpdateGeneration(completion=last_response, endTime=datetime.now(), usage=LlmUsage(**response.llm_output["token_usage"])))
        except Exception as e:
            logging.error(e)

    def on_llm_error(
        self,
        error: Union[Exception, KeyboardInterrupt],
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        try:
            logging.debug(f"on llm error: {run_id}")
            self.runs[run_id].state = self.runs[run_id].state.update(UpdateGeneration(endTime=datetime.now(), statusMessage=str(error), level=ObservationLevel.ERROR))
        except Exception as e:
            logging.error(e)

    def __join_tags_and_metadata(
        self,
        tags: Optional[List[str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> Optional[Dict[str, Any]]:
        if tags is None and metadata is None:
            return None
        elif tags is not None and len(tags) > 0:
            final_dict = {"tags": tags}
            if metadata is not None:
                final_dict.update(metadata)  # Merge metadata into final_dict
            return final_dict
        else:
            return metadata



import atexit
from datetime import datetime, timedelta
from enum import Enum
import logging
import queue
import threading


class Task:
    def __init__(self, task_id, function, predecessor_id: str = None):
        self.task_id = task_id
        self.predecessor_id = predecessor_id
        self.function = function
        self.result = None
        self.timestamp = None
        self.lock = threading.Lock()
        self.status = TaskStatus.UNSCHEDULED


class TaskStatus(Enum):
    SUCCESS = "success"
    FAIL = "fail"
    UNSCHEDULED = "unscheduled"


class TaskManager:
    def __init__(self, max_task_queue_size=10_000, max_task_age=600):
        self.max_task_queue_size = max_task_queue_size
        self.queue = queue.Queue(max_task_queue_size)
        self.consumer_thread = None
        self.result_mapping = {}
        self.max_task_age = max_task_age
        self.init_resources()

        # cleans up when the python interpreter closes
        atexit.register(self.join)

    def init_resources(self):
        self.consumer_thread = Consumer(self.queue, self.result_mapping, self.max_task_age)
        self.consumer_thread.start()

    def add_task(self, task_id, function, predecessor_id=None):
        try:
            logging.info(f"Adding task {task_id} with predecessor {predecessor_id}")
            if self.consumer_thread is None or not self.consumer_thread.is_alive():
                self.init_resources()
            task = Task(task_id, function, predecessor_id)

            self.queue.put(task, block=False)
            logging.info(f"Task {task_id} added to queue")
        except queue.Full:
            logging.warning("analytics-python queue is full")
            return False
        except Exception as e:
            logging.error(f"Exception in adding task {task_id} {e}")
            return False

    def flush(self):
        """Forces a flush from the internal queue to the server"""
        logging.info("flushing queue")
        queue = self.queue
        size = queue.qsize()
        queue.join()
        # Note that this message may not be precise, because of threading.
        logging.info("successfully flushed about %s items.", size)

    def join(self):
        """Ends the consumer thread once the queue is empty.
        Blocks execution until finished
        """
        logging.info("joining consumer thread")
        self.consumer_thread.pause()
        try:
            self.consumer_thread.join()
        except RuntimeError:
            # consumer thread has not started
            pass
        logging.info("consumer thread joined")

    def shutdown(self):
        """Flush all messages and cleanly shutdown the client"""
        logging.info("shutdown initiated")
        self.flush()
        self.join()
        logging.info("shutdown completed")

    def get_result(self, task_id):
        try:
            return self.result_mapping.get(task_id)
        except Exception as e:
            logging.error(f"Exception in getting result for task {task_id} {e}")


class Consumer(threading.Thread):
    def __init__(self, queue, result_mapping, max_task_age):
        """Create a consumer thread."""

        threading.Thread.__init__(self)
        # Make consumer a daemon thread so that it doesn't block program exit
        self.daemon = True
        self.queue = queue
        # It's important to set running in the constructor: if we are asked to
        # pause immediately after construction, we might set running to True in
        # run() *after* we set it to False in pause... and keep running
        # forever.
        self.running = True
        self.result_mapping = result_mapping
        self.max_task_age = max_task_age

    def run(self):
        """Runs the consumer."""
        logging.info("consumer is running...")
        while self.running:
            try:
                logging.info("consumer looping")
                self._prune_old_tasks(self.max_task_age)

                # elapsed = time.monotonic.monotonic() - start_time
                task = self.queue.get(block=True, timeout=1)

                self.result_mapping[task.task_id] = task

                logging.info(f"Task {task.task_id} received from the queue")

                self._execute_task(task)
                logging.info(f"Task {task.task_id} done")
                self.queue.task_done()

            except queue.Empty:
                break

        logging.debug("consumer exited.")

    def pause(self):
        """Pause the consumer."""
        self.running = False

    def _execute_task(self, task: Task):
        try:
            logging.info(f"Task {task.task_id} executing")

            result = None
            with task.lock:
                try:
                    result = task.function()
                    self.result_mapping[task.task_id].result = result
                    self.result_mapping[task.task_id].status = TaskStatus.SUCCESS
                    self.result_mapping[task.task_id].timestamp = datetime.now()
                    logging.info(f"Task {task.task_id} done with result {result}")
                except Exception as e:
                    self.result_mapping[task.task_id].result = e
                    self.result_mapping[task.task_id].status = TaskStatus.FAIL
                    self.result_mapping[task.task_id].timestamp = datetime.now()
                    logging.info(f"Task {task.task_id} failed with exception {e} ")
        except Exception as e:
            logging.error(f"Exception in the task {task.task_id} {e}")

    def _prune_old_tasks(self, delta: int):
        try:
            logging.info("Pruning old tasks")
            now = datetime.now()

            to_remove = [task_id for task_id, task in self.result_mapping.items() if task.status in [TaskStatus.SUCCESS, TaskStatus.FAIL] and task.timestamp and now - task.timestamp > timedelta(seconds=delta)]
            for task_id in to_remove:
                self.result_mapping.pop(task_id, None)
                logging.info(f"Task {task_id} pruned due to age")
        except Exception as e:
            logging.error(f"Exception in pruning old tasks {e}")



# This file was auto-generated by Fern from our API Definition.

import typing
import urllib.parse
from json.decoder import JSONDecodeError

import httpx
import pydantic

from ...core.api_error import ApiError
from ...core.jsonable_encoder import jsonable_encoder
from ...core.remove_none_from_headers import remove_none_from_headers
from ..commons.errors.access_denied_error import AccessDeniedError
from ..commons.errors.error import Error
from ..commons.errors.method_not_allowed_error import MethodNotAllowedError
from ..commons.errors.not_found_error import NotFoundError
from ..commons.errors.unauthorized_error import UnauthorizedError
from ..commons.types.score import Score
from .types.create_score_request import CreateScoreRequest
from .types.scores import Scores


class ScoreClient:
    def __init__(
        self,
        *,
        environment: str,
        x_langfuse_sdk_name: typing.Optional[str] = None,
        x_langfuse_sdk_version: typing.Optional[str] = None,
        username: str,
        password: str,
    ):
        self._environment = environment
        self.x_langfuse_sdk_name = x_langfuse_sdk_name
        self.x_langfuse_sdk_version = x_langfuse_sdk_version
        self._username = username
        self._password = password

    def create(self, *, request: CreateScoreRequest) -> Score:
        _response = httpx.request(
            "POST",
            urllib.parse.urljoin(f"{self._environment}/", "api/public/scores"),
            json=jsonable_encoder(request),
            headers=remove_none_from_headers(
                {"X-Langfuse-Sdk-Name": self.x_langfuse_sdk_name, "X-Langfuse-Sdk-Version": self.x_langfuse_sdk_version}
            ),
            auth=(self._username, self._password)
            if self._username is not None and self._password is not None
            else None,
            timeout=60,
        )
        if 200 <= _response.status_code < 300:
            return pydantic.parse_obj_as(Score, _response.json())  # type: ignore
        if _response.status_code == 400:
            raise Error(pydantic.parse_obj_as(str, _response.json()))  # type: ignore
        if _response.status_code == 401:
            raise UnauthorizedError(pydantic.parse_obj_as(str, _response.json()))  # type: ignore
        if _response.status_code == 403:
            raise AccessDeniedError(pydantic.parse_obj_as(str, _response.json()))  # type: ignore
        if _response.status_code == 405:
            raise MethodNotAllowedError(pydantic.parse_obj_as(str, _response.json()))  # type: ignore
        if _response.status_code == 404:
            raise NotFoundError(pydantic.parse_obj_as(str, _response.json()))  # type: ignore
        try:
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)

    def get(
        self,
        *,
        page: typing.Optional[int] = None,
        limit: typing.Optional[int] = None,
        user_id: typing.Optional[str] = None,
        name: typing.Optional[str] = None,
    ) -> Scores:
        _response = httpx.request(
            "GET",
            urllib.parse.urljoin(f"{self._environment}/", "api/public/scores"),
            params={"page": page, "limit": limit, "userId": user_id, "name": name},
            headers=remove_none_from_headers(
                {"X-Langfuse-Sdk-Name": self.x_langfuse_sdk_name, "X-Langfuse-Sdk-Version": self.x_langfuse_sdk_version}
            ),
            auth=(self._username, self._password)
            if self._username is not None and self._password is not None
            else None,
            timeout=60,
        )
        if 200 <= _response.status_code < 300:
            return pydantic.parse_obj_as(Scores, _response.json())  # type: ignore
        if _response.status_code == 400:
            raise Error(pydantic.parse_obj_as(str, _response.json()))  # type: ignore
        if _response.status_code == 401:
            raise UnauthorizedError(pydantic.parse_obj_as(str, _response.json()))  # type: ignore
        if _response.status_code == 403:
            raise AccessDeniedError(pydantic.parse_obj_as(str, _response.json()))  # type: ignore
        if _response.status_code == 405:
            raise MethodNotAllowedError(pydantic.parse_obj_as(str, _response.json()))  # type: ignore
        if _response.status_code == 404:
            raise NotFoundError(pydantic.parse_obj_as(str, _response.json()))  # type: ignore
        try:
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)


class AsyncScoreClient:
    def __init__(
        self,
        *,
        environment: str,
        x_langfuse_sdk_name: typing.Optional[str] = None,
        x_langfuse_sdk_version: typing.Optional[str] = None,
        username: str,
        password: str,
    ):
        self._environment = environment
        self.x_langfuse_sdk_name = x_langfuse_sdk_name
        self.x_langfuse_sdk_version = x_langfuse_sdk_version
        self._username = username
        self._password = password

    async def create(self, *, request: CreateScoreRequest) -> Score:
        async with httpx.AsyncClient() as _client:
            _response = await _client.request(
                "POST",
                urllib.parse.urljoin(f"{self._environment}/", "api/public/scores"),
                json=jsonable_encoder(request),
                headers=remove_none_from_headers(
                    {
                        "X-Langfuse-Sdk-Name": self.x_langfuse_sdk_name,
                        "X-Langfuse-Sdk-Version": self.x_langfuse_sdk_version,
                    }
                ),
                auth=(self._username, self._password)
                if self._username is not None and self._password is not None
                else None,
                timeout=60,
            )
        if 200 <= _response.status_code < 300:
            return pydantic.parse_obj_as(Score, _response.json())  # type: ignore
        if _response.status_code == 400:
            raise Error(pydantic.parse_obj_as(str, _response.json()))  # type: ignore
        if _response.status_code == 401:
            raise UnauthorizedError(pydantic.parse_obj_as(str, _response.json()))  # type: ignore
        if _response.status_code == 403:
            raise AccessDeniedError(pydantic.parse_obj_as(str, _response.json()))  # type: ignore
        if _response.status_code == 405:
            raise MethodNotAllowedError(pydantic.parse_obj_as(str, _response.json()))  # type: ignore
        if _response.status_code == 404:
            raise NotFoundError(pydantic.parse_obj_as(str, _response.json()))  # type: ignore
        try:
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)

    async def get(
        self,
        *,
        page: typing.Optional[int] = None,
        limit: typing.Optional[int] = None,
        user_id: typing.Optional[str] = None,
        name: typing.Optional[str] = None,
    ) -> Scores:
        async with httpx.AsyncClient() as _client:
            _response = await _client.request(
                "GET",
                urllib.parse.urljoin(f"{self._environment}/", "api/public/scores"),
                params={"page": page, "limit": limit, "userId": user_id, "name": name},
                headers=remove_none_from_headers(
                    {
                        "X-Langfuse-Sdk-Name": self.x_langfuse_sdk_name,
                        "X-Langfuse-Sdk-Version": self.x_langfuse_sdk_version,
                    }
                ),
                auth=(self._username, self._password)
                if self._username is not None and self._password is not None
                else None,
                timeout=60,
            )
        if 200 <= _response.status_code < 300:
            return pydantic.parse_obj_as(Scores, _response.json())  # type: ignore
        if _response.status_code == 400:
            raise Error(pydantic.parse_obj_as(str, _response.json()))  # type: ignore
        if _response.status_code == 401:
            raise UnauthorizedError(pydantic.parse_obj_as(str, _response.json()))  # type: ignore
        if _response.status_code == 403:
            raise AccessDeniedError(pydantic.parse_obj_as(str, _response.json()))  # type: ignore
        if _response.status_code == 405:
            raise MethodNotAllowedError(pydantic.parse_obj_as(str, _response.json()))  # type: ignore
        if _response.status_code == 404:
            raise NotFoundError(pydantic.parse_obj_as(str, _response.json()))  # type: ignore
        try:
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)



from enum import Enum
import logging
import os
import traceback
from typing import Optional
import uuid
from langfuse.api.resources.commons.types.create_event_request import CreateEventRequest
from langfuse.api.resources.commons.types.create_generation_request import CreateGenerationRequest
from langfuse.api.resources.commons.types.create_span_request import CreateSpanRequest
from langfuse.api.resources.score.types.create_score_request import CreateScoreRequest
from langfuse.api.resources.trace.types.create_trace_request import CreateTraceRequest
from langfuse.model import (
    CreateEvent,
    CreateGeneration,
    CreateScore,
    CreateSpan,
    CreateTrace,
    InitialGeneration,
    InitialScore,
    InitialSpan,
    UpdateGeneration,
    UpdateSpan,
)
from langfuse.api.resources.generations.types.update_generation_request import UpdateGenerationRequest
from langfuse.api.resources.span.types.update_span_request import UpdateSpanRequest
from langfuse.api.client import FintoLangfuse
from langfuse.task_manager import TaskManager
from .version import __version__ as version


class Langfuse:
    def __init__(self, public_key: str, secret_key: str, host: Optional[str] = None, release: Optional[str] = None):
        self.task_manager = TaskManager()

        self.base_url = host if host else "https://cloud.langfuse.com"

        self.client = FintoLangfuse(
            environment=self.base_url,
            username=public_key,
            password=secret_key,
            x_langfuse_sdk_name="python",
            x_langfuse_sdk_version=version,
        )

        self.trace_id = None

        self.release = os.environ.get("LANGFUSE_RELEASE", release)

    def get_trace_id(self):
        return self.trace_id

    def trace(self, body: CreateTrace):
        try:
            new_id = str(uuid.uuid4()) if body.id is None else body.id
            self.trace_id = new_id

            def task():
                try:
                    new_body = body.copy(update={"id": new_id})

                    if self.release is not None:
                        new_body = new_body.copy(update={"release": self.release})

                    logging.info(f"Creating trace {new_body}...")
                    return self.client.trace.create(request=new_body)
                except Exception as e:
                    traceback.print_exception(e)
                    raise e

            self.task_manager.add_task(new_id, task)

            return StatefulClient(self.client, new_id, StateType.TRACE, new_id, self.task_manager)
        except Exception as e:
            traceback.print_exception(e)

    def score(self, body: InitialScore):
        try:
            new_id = str(uuid.uuid4())

            def task():
                try:
                    new_body = body.copy(update={"id": new_id})
                    logging.info(f"Creating score {new_body}...")
                    return self.client.score.create(request=new_body)
                except Exception as e:
                    traceback.print_exception(e)
                    raise e

            self.task_manager.add_task(new_id, task)

            if body.observation_id is not None:
                return StatefulClient(self.client, body.observation_id, StateType.OBSERVATION, body.trace_id, self.task_manager)
            else:
                return StatefulClient(self.client, new_id, StateType.TRACE, new_id, self.task_manager)

        except Exception as e:
            traceback.print_exception(e)

    def span(self, body: InitialSpan):
        try:
            new_trace_id = str(uuid.uuid4()) if body.trace_id is None else body.trace_id
            self.trace_id = new_trace_id
            new_span_id = str(uuid.uuid4())

            def create_trace():
                try:
                    new_body = {
                        "id": new_trace_id,
                        "release": self.release,
                        "name": body.name,
                    }

                    logging.info(f"Creating trace {new_body}...")
                    request = CreateTraceRequest(**new_body)
                    return self.client.trace.create(request=request)
                except Exception as e:
                    traceback.print_exception(e)
                    raise e

            def create_span():
                try:
                    new_body = body.copy(update={"id": new_span_id})
                    new_body = body.copy(update={"trace_id": new_trace_id})
                    if self.release is not None:
                        new_body = body.copy(update={"trace": {"release": self.release}})
                    logging.info(f"Creating span {new_body}...")
                    request = CreateSpanRequest(**new_body.dict())
                    return self.client.span.create(request=request)
                except Exception as e:
                    traceback.print_exception(e)
                    raise e

            self.task_manager.add_task(new_trace_id, create_trace)
            self.task_manager.add_task(new_span_id, create_span)

            return StatefulSpanClient(self.client, new_span_id, StateType.OBSERVATION, new_trace_id, self.task_manager)
        except Exception as e:
            traceback.print_exception(e)

    def generation(self, body: InitialGeneration):
        try:
            new_trace_id = str(uuid.uuid4()) if body.trace_id is None else body.trace_id
            new_generation_id = str(uuid.uuid4())
            self.trace_id = new_trace_id
            print("new_trace_id", new_trace_id)

            def create_trace():
                try:
                    new_body = {
                        "id": new_trace_id,
                        "release": self.release,
                        "name": body.name,
                    }

                    logging.info(f"Creating trace {new_body}...")
                    request = CreateTraceRequest(**new_body)
                    return self.client.trace.create(request=request)
                except Exception as e:
                    traceback.print_exception(e)
                    raise e

            def create_generation():
                try:
                    new_body = body.copy(update={"id": new_generation_id})
                    new_body = body.copy(update={"trace_id": new_trace_id})
                    if self.release is not None:
                        new_body = body.copy(update={"trace": {"release": self.release}})
                    logging.info(f"Creating top-level generation {new_body}...")
                    request = CreateGenerationRequest(**new_body.dict())
                    return self.client.generations.log(request=request)
                except Exception as e:
                    traceback.print_exception(e)
                    raise e

            self.task_manager.add_task(new_generation_id, create_generation)
            self.task_manager.add_task(new_trace_id, create_trace)

            return StatefulGenerationClient(self.client, new_generation_id, StateType.OBSERVATION, new_trace_id, self.task_manager)
        except Exception as e:
            traceback.print_exception(e)

    # On program exit, allow the consumer thread to exit cleanly.
    # This prevents exceptions and a messy shutdown when the
    # interpreter is destroyed before the daemon thread finishes
    # execution. However, it is *not* the same as flushing the queue!
    # To guarantee all messages have been delivered, you'll still need
    # to call flush().
    def join(self):
        try:
            return self.task_manager.join()
        except Exception as e:
            traceback.print_exception(e)

    def flush(self):
        try:
            return self.task_manager.flush()
        except Exception as e:
            traceback.print_exception(e)

    def shutdown(self):
        try:
            return self.task_manager.shutdown()
        except Exception as e:
            traceback.print_exception(e)


class StateType(Enum):
    OBSERVATION = 1
    TRACE = 0


class StatefulClient:
    def __init__(self, client: Langfuse, id: str, state_type: StateType, trace_id: str, task_manager: TaskManager):
        self.client = client
        self.trace_id = trace_id
        self.id = id
        self.state_type = state_type
        self.task_manager = task_manager

    def _add_state_to_observation(self, body: dict):
        if self.state_type == StateType.OBSERVATION:
            body["parent_observation_id"] = self.id
            body["trace_id"] = self.trace_id
        else:
            body["trace_id"] = self.id
        return body

    def generation(self, body: CreateGeneration):
        try:
            generation_id = str(uuid.uuid4()) if body.id is None else body.id

            def task():
                try:
                    new_body = body.copy(update={"id": generation_id})

                    new_dict = self._add_state_to_observation(new_body.dict())

                    logging.info(f"Creating generation {new_dict}...")

                    request = CreateGenerationRequest(**new_dict)
                    return self.client.generations.log(request=request)
                except Exception as e:
                    traceback.print_exception(e)
                    raise e

            self.task_manager.add_task(generation_id, task)
        except Exception as e:
            traceback.print_exception(e)

        return StatefulGenerationClient(self.client, generation_id, StateType.OBSERVATION, self.trace_id, task_manager=self.task_manager)

    def span(self, body: CreateSpan):
        try:
            span_id = str(uuid.uuid4()) if body.id is None else body.id

            def task():
                try:
                    new_body = body.copy(update={"id": span_id})
                    logging.info(f"Creating span {new_body}...")

                    new_dict = self._add_state_to_observation(new_body.dict())

                    request = CreateSpanRequest(**new_dict)
                    return self.client.span.create(request=request)
                except Exception as e:
                    traceback.print_exception(e)
                    raise e

            self.task_manager.add_task(span_id, task)

            return StatefulSpanClient(self.client, span_id, StateType.OBSERVATION, self.trace_id, task_manager=self.task_manager)
        except Exception as e:
            traceback.print_exception(e)

    def score(self, body: CreateScore):
        try:
            score_id = str(uuid.uuid4()) if body.id is None else body.id

            def task():
                try:
                    new_body = body.copy(update={"id": score_id})
                    logging.info(f"Creating score {new_body}...")

                    new_dict = self._add_state_to_observation(new_body.dict())

                    if self.state_type == StateType.OBSERVATION:
                        new_dict["observationId"] = self.id

                    request = CreateScoreRequest(**new_dict)
                    return self.client.score.create(request=request)
                except Exception as e:
                    traceback.print_exception(e)
                    raise e

            self.task_manager.add_task(score_id, task)

            return StatefulClient(self.client, self.id, StateType.OBSERVATION, self.trace_id, task_manager=self.task_manager)
        except Exception as e:
            traceback.print_exception(e)

    def event(self, body: CreateEvent):
        try:
            event_id = str(uuid.uuid4()) if body.id is None else body.id

            def task():
                try:
                    new_body = body.copy(update={"id": event_id})
                    logging.info(f"Creating event {new_body}...")

                    new_dict = self._add_state_to_observation(new_body.dict())

                    request = CreateEventRequest(**new_dict)
                    return self.client.event.create(request=request)
                except Exception as e:
                    traceback.print_exception(e)
                    raise e

            self.task_manager.add_task(body.id, task)

            return StatefulClient(self.client, event_id, self.state_type, self.trace_id, task_manager=self.task_manager)
        except Exception as e:
            traceback.print_exception(e)


class StatefulGenerationClient(StatefulClient):
    def __init__(self, client: Langfuse, id: str, state_type: StateType, trace_id: str, task_manager: TaskManager):
        super().__init__(client, id, state_type, trace_id, task_manager)

    def update(self, body: UpdateGeneration):
        try:
            update_id = str(uuid.uuid4())

            def task():
                try:
                    new_body = body.copy(update={"generation_id": self.id})
                    logging.info(f"Update generation {new_body}...")
                    request = UpdateGenerationRequest(**new_body.dict())
                    return self.client.generations.update(request=request)
                except Exception as e:
                    traceback.print_exception(e)
                    raise e

            self.task_manager.add_task(update_id, task)

            return StatefulGenerationClient(self.client, self.id, StateType.OBSERVATION, self.trace_id, task_manager=self.task_manager)
        except Exception as e:
            traceback.print_exception(e)


class StatefulSpanClient(StatefulClient):
    def __init__(self, client: Langfuse, id: str, state_type: StateType, trace_id: str, task_manager: TaskManager):
        super().__init__(client, id, state_type, trace_id, task_manager)

    def update(self, body: UpdateSpan):
        try:
            update_id = str(uuid.uuid4())

            def task():
                try:
                    new_body = body.copy(update={"span_id": self.id})
                    logging.info(f"Update span {new_body}...")
                    request = UpdateSpanRequest(**new_body.dict())
                    return self.client.span.update(request=request)
                except Exception as e:
                    traceback.print_exception(e)
                    raise e

            self.task_manager.add_task(update_id, task)

            return StatefulSpanClient(self.client, self.id, StateType.OBSERVATION, self.trace_id, task_manager=self.task_manager)
        except Exception as e:
            traceback.print_exception(e)



# This file was auto-generated by Fern from our API Definition.

import enum
import typing

T_Result = typing.TypeVar("T_Result")


class TraceIdTypeEnum(str, enum.Enum):
    LANGFUSE = "LANGFUSE"
    EXTERNAL = "EXTERNAL"

    def visit(self, langfuse: typing.Callable[[], T_Result], external: typing.Callable[[], T_Result]) -> T_Result:
        if self is TraceIdTypeEnum.LANGFUSE:
            return langfuse()
        if self is TraceIdTypeEnum.EXTERNAL:
            return external()



# This file was auto-generated by Fern from our API Definition.

import datetime as dt
import typing

import pydantic

from ....core.datetime_utils import serialize_datetime
from .observation_level import ObservationLevel
from .trace_id_type_enum import TraceIdTypeEnum


class CreateEventRequest(pydantic.BaseModel):
    id: typing.Optional[str]
    trace_id: typing.Optional[str] = pydantic.Field(alias="traceId")
    trace_id_type: typing.Optional[TraceIdTypeEnum] = pydantic.Field(alias="traceIdType")
    name: typing.Optional[str]
    start_time: typing.Optional[dt.datetime] = pydantic.Field(alias="startTime")
    metadata: typing.Optional[typing.Any]
    input: typing.Optional[typing.Any]
    output: typing.Optional[typing.Any]
    level: typing.Optional[ObservationLevel]
    status_message: typing.Optional[str] = pydantic.Field(alias="statusMessage")
    parent_observation_id: typing.Optional[str] = pydantic.Field(alias="parentObservationId")
    version: typing.Optional[str]

    def json(self, **kwargs: typing.Any) -> str:
        kwargs_with_defaults: typing.Any = {"by_alias": True, "exclude_unset": True, **kwargs}
        return super().json(**kwargs_with_defaults)

    def dict(self, **kwargs: typing.Any) -> typing.Dict[str, typing.Any]:
        kwargs_with_defaults: typing.Any = {"by_alias": True, "exclude_unset": True, **kwargs}
        return super().dict(**kwargs_with_defaults)

    class Config:
        frozen = True
        allow_population_by_field_name = True
        json_encoders = {dt.datetime: serialize_datetime}
